﻿///////////////////////////////////////////////////////////////////////////
//  Copyright (C) Wizardry and Steamworks 2016 - License: GNU GPLv3      //
//  Please see: http://www.gnu.org/licenses/gpl.html for legal details,  //
//  rights of fair usage, the disclaimer and warranty conditions.        //
///////////////////////////////////////////////////////////////////////////

using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Threading;

namespace wasSharp.Collections.Specialized
{
    ///////////////////////////////////////////////////////////////////////////
    //    Copyright (C) 2016 Wizardry and Steamworks - License: GNU GPLv3    //
    ///////////////////////////////////////////////////////////////////////////
    /// <summary>
    ///     An implementation of an observable Dictionary.
    /// </summary>
    /// <typeparam name="K">the key type</typeparam>
    /// <typeparam name="V">the value type</typeparam>
    public class ObservableDictionary<K, V> : IDictionary<K, V>, INotifyCollectionChanged
    {
        private readonly ReaderWriterLockSlim SyncRoot = new ReaderWriterLockSlim(LockRecursionPolicy.SupportsRecursion);
        private readonly Dictionary<K, V> store = new Dictionary<K, V>();

        public bool IsVirgin { get; private set; } = true;

        public V this[K key]
        {
            get
            {
                SyncRoot.EnterReadLock();
                var v = store[key];
                SyncRoot.ExitReadLock();
                return v;
            }

            set
            {
                SyncRoot.EnterWriteLock();
                store[key] = value;
                SyncRoot.ExitWriteLock();
            }
        }

        public int Count => store.Count;

        public bool IsReadOnly => false;

        public ICollection<K> Keys
        {
            get
            {
                SyncRoot.EnterReadLock();
                var v = store.Keys;
                SyncRoot.ExitReadLock();
                return v;
            }
        }

        public ICollection<V> Values
        {
            get
            {
                SyncRoot.EnterReadLock();
                var v = store.Values;
                SyncRoot.ExitReadLock();
                return v;
            }
        }

        public void Add(KeyValuePair<K, V> item)
        {
            SyncRoot.EnterWriteLock();
            ((IDictionary<K, V>)store).Add(item);
            SyncRoot.ExitWriteLock();
            IsVirgin = false;
            OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, item));
        }

        public void Add(K key, V value)
        {
            SyncRoot.EnterWriteLock();
            store.Add(key, value);
            SyncRoot.ExitWriteLock();
            IsVirgin = false;
            OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add,
                new KeyValuePair<K, V>(key, value)));
        }

        public void Clear()
        {
            SyncRoot.EnterWriteLock();
            store.Clear();
            SyncRoot.ExitWriteLock();
            if (!IsVirgin)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
            IsVirgin = false;
        }

        public bool Contains(KeyValuePair<K, V> item)
        {
            SyncRoot.EnterReadLock();
            var c = ((IDictionary<K, V>)store).Contains(item);
            SyncRoot.ExitReadLock();
            return c;
        }

        public bool ContainsKey(K key)
        {
            SyncRoot.EnterReadLock();
            var c = store.ContainsKey(key);
            SyncRoot.ExitReadLock();
            return c;
        }

        public void CopyTo(KeyValuePair<K, V>[] array, int arrayIndex)
        {
            SyncRoot.EnterReadLock();
            ((IDictionary<K, V>)store).CopyTo(array, arrayIndex);
            SyncRoot.ExitReadLock();
        }

        public IEnumerator<KeyValuePair<K, V>> GetEnumerator()
        {
            SyncRoot.EnterReadLock();
            var enumerator = ((IDictionary<K, V>)store).GetEnumerator();
            SyncRoot.ExitReadLock();
            return enumerator;
        }

        public bool Remove(KeyValuePair<K, V> item)
        {
            SyncRoot.EnterWriteLock();
            var removed = ((IDictionary<K, V>)store).Remove(item);
            SyncRoot.ExitWriteLock();
            IsVirgin = false;
            if (removed)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, item));
            return removed;
        }

        public bool Remove(K key)
        {
            KeyValuePair<K, V> item;
            SyncRoot.EnterWriteLock();
            if (store.ContainsKey(key))
                item = new KeyValuePair<K, V>(key, store[key]);

            var removed = store.Remove(key);
            SyncRoot.ExitWriteLock();
            IsVirgin = false;
            if (removed)
                OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, item));
            return removed;
        }

        public bool TryGetValue(K key, out V value)
        {
            SyncRoot.EnterReadLock();
            var c = store.TryGetValue(key, out value);
            SyncRoot.ExitReadLock();
            return c;
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            SyncRoot.EnterReadLock();
            var enumerator = ((IDictionary<K, V>)store).GetEnumerator();
            SyncRoot.ExitReadLock();
            return enumerator;
        }

        public event NotifyCollectionChangedEventHandler CollectionChanged;

        private void OnCollectionChanged(NotifyCollectionChangedEventArgs args)
        {
            CollectionChanged?.Invoke(this, args);
        }
    }
}
